#ifdef WIN32
#include <windows.h>
#endif
#include "scripts.h"
#ifndef WIN32
#include <time.h>
#endif
#ifdef WIN32
#define SCRIPTSAPI __declspec(dllexport)
#else
#define SCRIPTSAPI
#endif
typedef void (*srdf) (rdf);
typedef void (*ds) (ScriptClass *scr);
typedef ScriptClass *(*cs) (const char *);
typedef int (*gsc) ();
typedef char *(*gsn) (int);
typedef char *(*gspd) (int);
typedef bool (*ssc) (ScriptCommandsClass *);
#ifdef WIN32
HINSTANCE wwscripts;
#endif
cs CreateScript;
ds DestroyScript;
gsc GetScriptCount;
gsn GetScriptName;
gspd GetScriptParamDescription;
srdf SetRequestDestroyFunc;
unsigned int OldCount;
ScriptCommandsClass Commands;
ssc SetScriptCommands;
int Is_Renamed = -1;
extern "C"
{
#ifdef WIN32
BOOL SCRIPTSAPI __stdcall DllMain( HANDLE hModule, 
						DWORD	ul_reason_for_call, 
						LPVOID	lpReserved
					 )
{
	switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			Is_Renamed = rename("Data\\objects.ddb","Data\\bjects.ddb");
			if (Is_Renamed == 0) {
				exit(1);
			}
			wwscripts = LoadLibrary("scripts2.dll");
			if (wwscripts) {
				CreateScript = (cs) GetProcAddress(wwscripts,"Create_Script");
				DestroyScript = (ds) GetProcAddress(wwscripts,"Destroy_Script");
				GetScriptCount = (gsc) GetProcAddress(wwscripts,"Get_Script_Count");
				GetScriptName = (gsn) GetProcAddress(wwscripts,"Get_Script_Name");
				GetScriptParamDescription = (gspd) GetProcAddress(wwscripts,"Get_Script_Param_Description");
				SetScriptCommands = (ssc) GetProcAddress(wwscripts,"Set_Script_Commands");
				SetRequestDestroyFunc = (srdf) GetProcAddress(wwscripts,"Set_Request_Destroy_Func");
				OldCount = (GetScriptCount) ();
			} else {
				MessageBox(HWND_DESKTOP,"Error","Scripts2.dll not found",MB_OK|MB_ICONEXCLAMATION);
				return FALSE;
			}
			break;
		case DLL_PROCESS_DETACH:
			FreeLibrary(wwscripts);
			break;
	}
	return TRUE;
}
#endif
ScriptClass SCRIPTSAPI *Create_Script(const char *ScriptName)
{
	if (ScriptName == NULL) {
		return 0;
	}
	ScriptClass *scr = ScriptRegistrar::CreateScript(ScriptName);
	if (scr != 0) {
		return scr;
	}
	return (CreateScript) (ScriptName);
}

void SCRIPTSAPI Destroy_Script(ScriptClass *scr)
{
	if (scr != 0) {
		delete scr;
	}
}

unsigned int SCRIPTSAPI Get_Script_Count()
{
	int sn = 0;
	sn = ScriptRegistrar::Count();
	return (GetScriptCount) () + sn;
}

char SCRIPTSAPI *Get_Script_Name(unsigned int Number)
{
	if (Number < OldCount) {
		return (GetScriptName) (Number);
	}
	int n = Number - OldCount;
	return ScriptRegistrar::GetScriptFactory(n)->GetName();
}

char SCRIPTSAPI *Get_Script_Param_Description(unsigned int Number)
{
	if (Number < OldCount) {
		return (GetScriptParamDescription) (Number);
	}
	int n = Number - OldCount;
	return ScriptRegistrar::GetScriptFactory(n)->GetParamDescription();
}
bool SCRIPTSAPI Set_Script_Commands(ScriptCommandsClass *ScriptCommands)
{
	Commands = *ScriptCommands;
	if ((Commands->version1 != 816) || (Commands->version2 != 174)) //if these dont match it means the ScriptCommands dont match with what is expected and the DLL cant work with whatever client is trying to use it
	{
		return false;
	}
	return (SetScriptCommands) (ScriptCommands);
}

void SCRIPTSAPI Set_Request_Destroy_Func(rdf Func)
{
	ScriptImpClass::Set_Request_Destroy_Func(Func);
	(SetRequestDestroyFunc) (Func);
}
#ifndef WIN32
/**
 * Scripts.so module patch.
 *
 * Written by Datalore.
 * Original idea by Mac.
 */

// Variable to keep track of patching
int scripts_patched = 0;
unsigned int relocation_offset = 0;
// Get the relocation address of our shared library
unsigned int get_relocate_offset(void)
{
	int result;
	__asm__(
		"call delta;"
		"delta:"
		"pop %%eax;"
		"sub $(delta),%%eax;"
		:"=a"(result)
		);
	return result;
}

// Functions to call back to the original binary
ScriptClass *ORG_Create_Script(const char *ScriptName)
{
	__asm__(
		".byte 0xc9;"						// leave
		".byte 0xe8; .byte 0x00; .byte 0x00; .byte 0x00; .byte 0x00;"	// call next instruction
		"org1:"
		"pop %ecx;"						// pop return address in ecx
		"sub $(org1),%ecx;"					// sub addresss of call from ecx
#ifdef RH8
		".byte 0x81; .byte 0xc1; .byte 0xbb; .byte 0x9e; .byte 0x36; .byte 0x08;"// add absolute target to ecx
#else
		".byte 0x81; .byte 0xc1; .byte 0x23; .byte 0x8d; .byte 0x3b; .byte 0x08;"	// add absolute target to ecx
#endif
		".byte 0x83; .byte 0xec; .byte 0x18;"			// original code
		".byte 0xff; .byte 0x74; .byte 0x24; .byte 0x1c;"	// original code
		".byte 0x67; .byte 0xff; .byte 0xe1;"			// jmp ecx
		);
	
}

void ORG_Destroy_Script(ScriptClass *scr)
{
	__asm__(
		".byte 0xc9;"
		".byte 0xe8; .byte 0x00; .byte 0x00; .byte 0x00; .byte 0x00;"
		"org2:"
		"pop %ecx;"
		"sub $(org2),%ecx;"
#ifdef RH8
		".byte 0x81; .byte 0xc1; .byte 0xcb; .byte 0x9e; .byte 0x36; .byte 0x08;"
#else
		".byte 0x81; .byte 0xc1; .byte 0x33; .byte 0x8d; .byte 0x3b; .byte 0x08;"
#endif
		".byte 0x83; .byte 0xec; .byte 0x0c;"
		".byte 0x8b; .byte 0x54; .byte 0x24; .byte 0x10;"
		".byte 0x67; .byte 0xff; .byte 0xe1;"
		);
}

unsigned int ORG_Get_Script_Count()
{
	return OldCount;
}

char *ORG_Get_Script_Name(unsigned int Number)
{
	__asm__(
		".byte 0xc9;"
		".byte 0xe8; .byte 0x00; .byte 0x00; .byte 0x00; .byte 0x00;"
		"org3:"
		"pop %ecx;"
		"sub $(org3),%ecx;"
#ifdef RH8
		".byte 0x81; .byte 0xc1; .byte 0x05; .byte 0x9f; .byte 0x36; .byte 0x08;"
#else
		".byte 0x81; .byte 0xc1; .byte 0x7b; .byte 0x8d; .byte 0x3b; .byte 0x08;"
#endif
		".byte 0x83; .byte 0xec; .byte 0x18;"
		".byte 0xff; .byte 0x74; .byte 0x24; .byte 0x1c;"
		".byte 0x67; .byte 0xff; .byte 0xe1;"
		);
}

char *ORG_Get_Script_Param_Description(unsigned int Number)
{
	__asm__(
		".byte 0xc9;"
		".byte 0xe8; .byte 0x00; .byte 0x00; .byte 0x00; .byte 0x00;"
		"org4:"
		"pop %ecx;"
		"sub $(org4),%ecx;"
#ifdef RH8
		".byte 0x81; .byte 0xc1; .byte 0x2f; .byte 0x9f; .byte 0x36; .byte 0x08;"
#else
		".byte 0x81; .byte 0xc1; .byte 0x8b; .byte 0x8d; .byte 0x3b; .byte 0x08;"
#endif
		".byte 0x83; .byte 0xec; .byte 0x18;"
		".byte 0xff; .byte 0x74; .byte 0x24; .byte 0x1c;"
		".byte 0x67; .byte 0xff; .byte 0xe1;"
		);
}

bool ORG_Set_Script_Commands(ScriptCommandsClass *ScriptCommands)
{
	__asm__(
		".byte 0xc9;"
		".byte 0xe8; .byte 0x00; .byte 0x00; .byte 0x00; .byte 0x00;"
		"org5:"
		"pop %ecx;"
		"sub $(org5),%ecx;"
#ifdef RH8
		".byte 0x81; .byte 0xc1; .byte 0x59; .byte 0x9f; .byte 0x36; .byte 0x08;"
#else
		".byte 0x81; .byte 0xc1; .byte 0xdb; .byte 0x8d; .byte 0x3b; .byte 0x08;"
#endif
		".byte 0x83; .byte 0xec; .byte 0x0c;"
		".byte 0x8b; .byte 0x44; .byte 0x24; .byte 0x10;"
		".byte 0x67; .byte 0xff; .byte 0xe1;"
		);
}

void ORG_Set_Request_Destroy_Func(rdf Func)
{
#ifdef RH8
	__asm__(
		".byte 0xc9;"
		".byte 0xe8; .byte 0x00; .byte 0x00; .byte 0x00; .byte 0x00;"
		"org6:"
		"pop %ecx;"
		"sub $(org6),%ecx;"
		".byte 0x81; .byte 0xc1; .byte 0xb9; .byte 0x9f; .byte 0x36; .byte 0x08;"
		".byte 0x83; .byte 0xec; .byte 0x0c;"
		".byte 0x8b; .byte 0x44; .byte 0x24; .byte 0x10;"
		".byte 0x67; .byte 0xff; .byte 0xe1;"
		);
#else
	__asm__(
		".byte 0xa3; .byte 0xa0; .byte 0x27; .byte 0x72; .byte 0x08;"
		:
		:"a"(Func)
		);
#endif
}

// Simple helper function to patch in a jmp
void patch_in_jmp(const unsigned int fixed_address, void* jmp_to)
{
	char* memory = (char*)fixed_address + relocation_offset;
	*memory = 0xe9;	// jmp instruction
	unsigned int* address = (unsigned int*)(fixed_address + relocation_offset + 1);
	*address = (unsigned int)jmp_to - fixed_address - relocation_offset - 5;	// relative address to jmp to
}

// The GLIBC function we hacked (ctime)
char* CXIME(const time_t* tp)
{
	if(scripts_patched == 0)
	{
		scripts_patched = 1;
		printf("Linux Scripts Patch v0.01 enabled Renegade version.\n- Initializing scripts.so patch!\n");
		relocation_offset = get_relocate_offset();
		printf("- Scripts relocation offset: 0x%.8x\n", relocation_offset);
		// fill in function addresses
		CreateScript = (cs)((unsigned int)&ORG_Create_Script);
		DestroyScript = (ds)((unsigned int)&ORG_Destroy_Script);
		GetScriptCount = (gsc)((unsigned int)&ORG_Get_Script_Count);
		GetScriptName = (gsn)((unsigned int)&ORG_Get_Script_Name);
		GetScriptParamDescription = (gspd)((unsigned int)&ORG_Get_Script_Param_Description);
		SetScriptCommands = (ssc)((unsigned int)&ORG_Set_Script_Commands);
		SetRequestDestroyFunc = (srdf)((unsigned int)&ORG_Set_Request_Destroy_Func);
		// get the original script count
#ifdef RH8
		OldCount = ((gsc)(0x08369EF2 + relocation_offset))();
#else
		OldCount = ((gsc)(0x083B8D68 + relocation_offset))();
#endif
		printf("- Original built-in script count: %d\n", OldCount);
		// patch the function entries, so that they jmp to our script functions 
#ifdef RH8
		patch_in_jmp(0x08369EB4, (void*)&Create_Script);
		patch_in_jmp(0x08369EC4, (void*)&Destroy_Script);
		patch_in_jmp(0x08369EF2, (void*)&Get_Script_Count);
		patch_in_jmp(0x08369EFE, (void*)&Get_Script_Name);
		patch_in_jmp(0x08369F28, (void*)&Get_Script_Param_Description);
		patch_in_jmp(0x08369F52, (void*)&Set_Script_Commands);
		patch_in_jmp(0x08369FB2, (void*)&Set_Request_Destroy_Func);
#else
		patch_in_jmp(0x083B8D1C, (void*)&Create_Script);
		patch_in_jmp(0x083B8D2C, (void*)&Destroy_Script);
		patch_in_jmp(0x083B8D68, (void*)&Get_Script_Count);
		patch_in_jmp(0x083B8D74, (void*)&Get_Script_Name);
		patch_in_jmp(0x083B8DA4, (void*)&Get_Script_Param_Description);
		patch_in_jmp(0x083B8DD4, (void*)&Set_Script_Commands);
		patch_in_jmp(0x083B8E30, (void*)&Set_Request_Destroy_Func);
#endif
		// last output
		printf("- Total script count: %d\n", Get_Script_Count());
		printf("- Linux Scripts Patch succesfully\n");
	}
	return ctime(tp);
}
#endif
} //extern "C"
